from __future__ import unicode_literals
from django.db import models
from django.db.models import Q
from django.utils.translation import ugettext_lazy as _
from django.conf import settings
from django.utils.encoding import python_2_unicode_compatible

class SearchFormManager(models.Manager):
    def get_for_user(self, user):
        if user.is_anonymous():
            return self.filter(public=True)

        return self.filter(
            Q(public=True) | Q(owner=user))


@python_2_unicode_compatible
class SearchForm(models.Model):
    name = models.TextField(verbose_name=_("Name"), unique=True)
    owner = models.ForeignKey(settings.AUTH_USER_MODEL)
    public = models.BooleanField(
        verbose_name=_("Public"),
        default=False, help_text=_(
            "Make this search publicly available?"))
    data = models.TextField(verbose_name=_("Form data (JSON)"))

    objects = SearchFormManager()

    class Meta:
        ordering = ['name',]

    def __str__(self):
        return self.name


# -*- encoding: utf-8 -*-
from __future__ import unicode_literals
from builtins import str as text


def make_field(klass, operation, value, prev_op="and"):
    return {
        'field': text(klass.label),
        'operator': text(operation),
        'value': value,
        'prev_op': prev_op}


# -*- encoding: utf-8 -*-

from django.conf.urls import  url
from django.conf import settings
from django.views.decorators.csrf import csrf_exempt
from django.views.i18n import javascript_catalog

from multiseek import views
from multiseek.views import load_form, MultiseekModelRouter

js_info_dict = {
    'domain': 'djangojs',
    'packages': ('multiseek',),
}

urlpatterns = [

    url(r'^jsi18n/$',
        javascript_catalog,
        js_info_dict,
        name="js_i18n"),

    url(r'^$', csrf_exempt(views.MultiseekFormPage.as_view(
        registry=settings.MULTISEEK_REGISTRY,
        template_name="multiseek/index.html"
    )), name="index"),

    url(r'^results/$',
        csrf_exempt(views.MultiseekResults.as_view(
            registry=settings.MULTISEEK_REGISTRY,
            template_name="multiseek/results.html"
        )), name="results"),

    url(r'^save_form/$',
        csrf_exempt(views.MultiseekSaveForm.as_view(
            registry=settings.MULTISEEK_REGISTRY
        )), name="save_form"),

    url(r'^reset/$',
        csrf_exempt(views.reset_form),
        name="reset"),

    url(r'^remove-from-results/(?P<pk>\d+)$',
        views.remove_by_hand,
        name="remove_from_results"),

    url(r'^remove-from-removed-results/(?P<pk>\d+)$',
        views.remove_from_removed_by_hand,
        name="remove_from_removed_results"),

    url(r'^reenable-removed-ids/$',
        views.reenable_removed_by_hand,
        name="reenable_removed_ids"),

   url(r'^load_form/(?P<search_form_pk>\d+)',
        load_form,
        name="load_form"),

    url(r'^autocomplete/(?P<model>.*)/$', MultiseekModelRouter.as_view(
        registry=settings.MULTISEEK_REGISTRY
    ))

]


# -*- encoding: utf-8 -*-
from __future__ import unicode_literals
import json
from unittest import TestCase

import pytest
from mock import MagicMock

from multiseek.logic import UnknownOperation, AutocompleteQueryObject, \
    RangeQueryObject, RANGE_OPS, StringQueryObject, QueryObject, DIFFERENT, \
    NOT_CONTAINS, NOT_STARTS_WITH, MultiseekRegistry, STRING, ParseError, \
    UnknownField, EQUALITY_OPS_ALL, OR, AND, create_registry, get_registry, \
    EQUAL, IntegerQueryObject, LESSER_OR_EQUAL, RANGE, ReportType, Ordering, MULTISEEK_ORDERING_PREFIX
from multiseek.models import SearchForm
from multiseek.util import make_field
import six
from builtins import str as text

test_json = json.dumps({'form_data': [None,
    dict(field='foo', operator=text(EQUALITY_OPS_ALL[0]), value='foo', prev_op=None)]})

def py3k_test_string(s):
    if six.PY3:
        return s.replace("u'", "'").replace('u"', '"').replace(", u'", ", '")
    return s

class TestQueryObject(TestCase):
    def setUp(self):
        class MyQueryObject(QueryObject):
            field_name = "foo"

        self.q = MyQueryObject()

    def test_value_from_web(self):
        self.assertEquals(123, self.q.value_from_web(123))


    def test_query_for(self):
        res = self.q.query_for("foobar", DIFFERENT)
        self.assertEquals(
            py3k_test_string("(NOT (AND: (u'foo', u'foobar')))"),
            str(res))

    def test_query_for_raises(self):
        self.assertRaises(
            UnknownOperation, self.q.query_for, 'foobar', 'unknown operation')


class TestStringQueryObject(TestCase):
    def setUp(self):
        class MyStringQueryObject(StringQueryObject):
            field_name = "foo"

        self.q = MyStringQueryObject()

    def test_query_for(self):
        args = [
            (DIFFERENT, "(NOT (AND: (u'foo', u'foobar')))"),
            (NOT_CONTAINS, "(NOT (AND: (u'foo__icontains', u'foobar')))"),
            (NOT_STARTS_WITH,
             "(NOT (AND: (u'foo__startswith', u'foobar')))")
        ]

        for param, result in args:
            res = self.q.real_query("foobar", param)
            self.assertEquals(
                str(res),
                py3k_test_string(result))

    def test_query_for_raises(self):
        self.assertRaises(
            UnknownOperation,
            self.q.real_query, 'lol', 'bad operation')


class TestAutocompleteQueryObject(TestCase):
    def test_value_from_web(self):
        q = AutocompleteQueryObject('foo')

        q.model = MagicMock()
        q.model.objects.get.return_value = True
        self.assertEquals(q.value_from_web(None), None)
        self.assertEquals(q.value_from_web('foo'), None)
        self.assertEquals(q.value_from_web(1), True)

    def test_value_to_web(self):
        q = AutocompleteQueryObject('foo')
        q.model = MagicMock()
        q.model.objects.get.return_value = True

        self.assertEquals(q.value_to_web(1), '[1, "True"]')

    @pytest.mark.django_db
    def test_value_to_web_bug(self):
        q = AutocompleteQueryObject('fo', model=SearchForm)
        self.assertEquals(q.value_to_web(1), '[null, ""]')

class TestRangeQueryObject(TestCase):
    def test_value_from_web(self):
        r = RangeQueryObject('foo')
        self.assertEquals(r.value_from_web("[1,2]"), [1, 2])
        self.assertEquals(r.value_from_web('["1","2"]'), [1, 2])
        self.assertEquals(r.value_from_web("[1,2,3]"), None)
        self.assertEquals(r.value_from_web('["foo","bar"]'), None)
        self.assertEquals(r.value_from_web('123'), None)

        self.assertRaises(
            UnknownOperation, r.real_query, [1, 2], 'foo')

        res = r.real_query([1, 2], RANGE_OPS[1])

        s1 = "(NOT (AND: (u'foo__gte', 1), (u'foo__lte', 2)))"
        maybe_this = str(res) == py3k_test_string(s1)

        s2 = "(NOT (AND: (u'foo__lte', 2), (u'foo__gte', 1)))"
        maybe_that = str(res) == py3k_test_string(s2)

        self.assert_(maybe_that or maybe_this)


class TestIntegerQueryObject(TestCase):
    def test_value_from_web(self):
        r = IntegerQueryObject('foo')
        self.assertEquals(r.value_from_web('123'), 123)

        self.assertRaises(
            UnknownOperation, r.real_query, 123, 'foo')

        res = r.real_query(123, text(LESSER_OR_EQUAL))
        self.assertEquals(
            str(res),
            py3k_test_string("(AND: (u'foo__lte', 123))"))


class TestMultiseekRegistry(TestCase):
    def setUp(self):
        self.registry = MultiseekRegistry()
        self.registry.add_field(StringQueryObject('foo'))
        self.registry.add_field(RangeQueryObject('bar'))
        self.registry.add_field(RangeQueryObject('quux', public=False))
        self.registry.report_types = [
            ReportType("list", "List"),
            ReportType("table", "Table")
        ]
        self.registry.ordering = [
            Ordering("foo", "Foo"),
            Ordering("bar", "Bar"),
        ]

    def test_add_field_raises(self):
        self.assertRaises(
            AssertionError, self.registry.add_field, StringQueryObject('foo'))

    def test_field_by_type(self):
        self.assertEquals(
            len(self.registry.field_by_type(STRING)),
            1)

        self.assertEquals(
            len(self.registry.field_by_type(RANGE, public=False)),
            2)

    def test_extract(self):
        self.assertEquals(
            self.registry.extract('field_name'), ['foo', 'bar'])

        self.assertEquals(
            self.registry.extract('field_name', public=False),
            ['foo', 'bar', 'quux'])

    def test_parse_field(self):
        self.assertRaises(
            ParseError,
            self.registry.parse_field, {})

        self.assertRaises(
            UnknownField,
            self.registry.parse_field,
            dict(field='XXX', operator='IS', value='LOL', prev_op=None))

        self.assertRaises(
            UnknownOperation,
            self.registry.parse_field,
            dict(field='foo', operator='XXX', value='FO', prev_op=None))

        res = self.registry.parse_field(
            dict(field='foo', operator=EQUALITY_OPS_ALL[0], value='foo', prev_op=None))

        self.assertEquals(str(res),
                          py3k_test_string("(AND: (u'foo', u'foo'))"))

    def test_get_recursive_list(self):
        input = [None,
            [None, [None, dict(field='foo', operator=text(EQUALITY_OPS_ALL[0]), value='foo', prev_op=None)]],
            dict(field='foo', operator=text(EQUALITY_OPS_ALL[0]), value='bar', prev_op="BAD OP")]

        self.assertRaises(
            UnknownOperation,
            self.registry.get_query_recursive,
            input)

        input[2]['prev_op'] = OR

        res = self.registry.get_query_recursive(input)
        self.assertEquals(
            str(res),
            py3k_test_string("(OR: (u'foo', u'foo'), (u'foo', u'bar'))"))

    def test_get_query(self):
        gq = self.registry.get_query(json.loads(test_json)['form_data'])
        self.assertEquals(
            str(gq),
            py3k_test_string("(AND: (u'foo', u'foo'))"))

    def test_get_query_for_model(self):
        self.registry.model = MagicMock()
        self.registry.get_query_for_model(json.loads(test_json))
        self.registry.get_query_for_model(None)

    def test_recreate_form(self):
        op = text(EQUALITY_OPS_ALL[0])
        fld_noop = dict(field='foo', operator=op, value=u'foo', prev_op=None)
        fld_and = dict(field='foo', operator=op, value=u'foo', prev_op=AND)
        fld_or = dict(field='foo', operator=op, value=u'foo', prev_op=OR)
        res = self.registry.recreate_form(
            {'form_data':
                 [None,
                  fld_noop,
                  fld_or, [
                     AND,
                     fld_noop,
                     fld_or],
                  fld_and,
                  [OR,
                   fld_noop,
                   fld_or]
                 ],
             'ordering': {
                 '%s1' % MULTISEEK_ORDERING_PREFIX: "1",
                 '%s1_dir' % MULTISEEK_ORDERING_PREFIX: "1",
             },
             'report_type': '1'})

        self.maxDiff = None

        ex = u"""$('#frame-0').multiseekFrame('addField', 'foo', 'equals', 'foo', null);
$('#frame-0').multiseekFrame('addField', 'foo', 'equals', 'foo', 'or');
$('#frame-0').multiseekFrame('addFrame', 'and');
$('#frame-1').multiseekFrame('addField', 'foo', 'equals', 'foo', null);
$('#frame-1').multiseekFrame('addField', 'foo', 'equals', 'foo', 'or');
$('#frame-0').multiseekFrame('addField', 'foo', 'equals', 'foo', 'and');
$('#frame-0').multiseekFrame('addFrame', 'or');
$('#frame-2').multiseekFrame('addField', 'foo', 'equals', 'foo', null);
$('#frame-2').multiseekFrame('addField', 'foo', 'equals', 'foo', 'or');
\t\t$("select[name=order_1] option").eq(1).prop("selected", true);
\t\t$("input[name=order_1_dir]").attr("checked", true);
\t\t$("select[name=_ms_report_type] option").eq(1).prop("selected", true);
\t\tif (window.Foundation) {
\t\t\t$("input[name=order_1_dir]").next().toggleClass("checked", true)
\t\t}\n""" % dict(equal=EQUAL)

        self.assertEquals(ex, res)

    def     test_create_registry(self):
        create_registry(None, StringQueryObject('foo'))

    def test_get_registry(self):
        r = get_registry('test_app.multiseek_registry')
        from test_app.multiseek_registry import registry

        self.assertEquals(r, registry)

    def test_bug_3(self):
        f = self.registry.fields[0]
        v = self.registry.fields[0].ops[0]
        value = 'foo'

        field = make_field(f, v, value)
        field_or = make_field(f, v, value, "lol")

        form = {'form_data': [None, field, field, field, field_or]}

        self.assertRaises(
            ParseError, self.registry.recreate_form, form)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='SearchForm',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.TextField(unique=True, verbose_name='Name')),
                ('public', models.BooleanField(default=False, help_text='Make this search publicly available?', verbose_name='Public')),
                ('data', models.TextField(verbose_name='Form data (JSON)')),
                ('owner', models.ForeignKey(to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
    ]



# -*- encoding: utf-8 -*-
from __future__ import unicode_literals

from decimal import Decimal
import decimal
import importlib

import json
import re
from datetime import timedelta, datetime
from dateutil.parser import parse
from django.db.models import Q
from django.utils import html
try:
    from django.db.models.options import get_verbose_name
except ImportError:
    from django.utils.text import camel_case_to_spaces as get_verbose_name
from django.utils.translation import ugettext_lazy as _
from collections import namedtuple
from builtins import str as text

import six
if six.PY3:
    unicode = str

MULTISEEK_REPORT_TYPE = '_ms_report_type'
MULTISEEK_ORDERING_PREFIX = "order_"

AND = "and"
OR = "or"
ANDNOT = "andnot"

STRING = "string"
INTEGER = "integer"
DECIMAL = "decimal"
AUTOCOMPLETE = "autocomplete"
RANGE = "range"
DATE = "date"
VALUE_LIST = 'value-list'

EQUAL = _("equals")         # u"równy"
EQUAL_FEMALE = _("equals (female geder)")  # u"równa"
EQUAL_NONE = _("equals (no gender)")    # u'równe'
EQUAL_BOTH = _("equals (both genders)")    #u'równy/a'

GREATER = _("greater, than")
GREATER_FEMALE = _("greater, than (female gender)")
GREATER_NONE = _("greater, than (no gender)")
GREATER_OPS_ALL = [GREATER, GREATER_FEMALE, GREATER_NONE]

LESSER = _("lesser, than")
LESSER_FEMALE = _("lesser, than (female gender)")
LESSER_NONE = _("lesser, than (no gender)")
LESSER_OPS_ALL = [LESSER, LESSER_FEMALE, LESSER_NONE]

GREATER_OR_EQUAL = _("greater or equal to")
GREATER_OR_EQUAL_FEMALE = _("greater or equal to(female gender)")
GREATER_OR_EQUAL_NONE = _("greater or equal to (no gender)")
GREATER_OR_EQUAL_OPS_ALL = [GREATER_OR_EQUAL, GREATER_OR_EQUAL_FEMALE,
                            GREATER_OR_EQUAL_NONE]

LESSER_OR_EQUAL = _("lesser or equal to")
LESSER_OR_EQUAL_FEMALE = _("lesser or equal to (female gender)")
LESSER_OR_EQUAL_NONE = _("lesser or equal to (no gender)")
LESSER_OR_EQUAL_OPS_ALL = [LESSER_OR_EQUAL, LESSER_OR_EQUAL_FEMALE,
                           LESSER_OR_EQUAL_NONE]

DIFFERENT = _("differs")        # u"różny"
DIFFERENT_FEMALE = _("differs (female gender)") # u'różna'
DIFFERENT_NONE = _("differs (no gender)")   # u'różne'
DIFFERENT_BOTH = _("differs (both gender)")   # u'różny/a'

CONTAINS = _("contains")        # u"zawiera"
NOT_CONTAINS = _("not contains")    # u"nie zawiera"
STARTS_WITH = _("starts with")  # u"zaczyna się od"
NOT_STARTS_WITH = _("not starts with")  # u"nie zaczyna się od"

IN_RANGE = _("in range")              # u'zawiera się w'
NOT_IN_RANGE = _("outside range")      # u'nie zawiera się w'

STRING_OPS = [CONTAINS, NOT_CONTAINS,
              EQUAL, DIFFERENT,
              STARTS_WITH, NOT_STARTS_WITH]

INTEGER_OPS_MALE = [EQUAL, DIFFERENT, GREATER, LESSER]
INTEGER_OPS_FEMALE = [EQUAL_FEMALE, DIFFERENT_FEMALE, GREATER_FEMALE,
                      LESSER_FEMALE]
INTEGER_OPS_NONE = [EQUAL_NONE, DIFFERENT_NONE, GREATER_NONE, LESSER_NONE]
INTEGER_OPS_ALL = INTEGER_OPS_FEMALE + INTEGER_OPS_MALE + INTEGER_OPS_NONE

RANGE_OPS = [IN_RANGE, NOT_IN_RANGE]
DATE_OPS = [EQUAL_FEMALE, DIFFERENT_FEMALE, GREATER_FEMALE,
            GREATER_OR_EQUAL_FEMALE, LESSER_FEMALE, LESSER_OR_EQUAL_FEMALE,
            IN_RANGE, NOT_IN_RANGE]
EQUALITY_OPS_MALE = [EQUAL, DIFFERENT]
EQUALITY_OPS_FEMALE = [EQUAL_FEMALE, DIFFERENT_FEMALE]
EQUALITY_OPS_NONE = [EQUAL_NONE, DIFFERENT_NONE]
EQUALITY_OPS_BOTH = [EQUAL_BOTH, DIFFERENT_BOTH]

EQUALITY_OPS_ALL = EQUALITY_OPS_MALE + EQUALITY_OPS_FEMALE + \
                   EQUALITY_OPS_NONE + EQUALITY_OPS_BOTH

DIFFERENT_ALL = DIFFERENT, DIFFERENT_FEMALE, DIFFERENT_NONE, DIFFERENT_BOTH


class UnknownOperation(Exception):
    pass


class UnknownField(Exception):
    pass


class ParseError(Exception):
    pass


class QueryObject(object):
    """This is a Query Object!

    This object replaces the parameters that it gets from the web UI, which
    is the multiseek main form (field, operation, value) to a list of
    Q objects that can be used to query the database with Django.
    """

    field_name = None
    label = None
    type = None
    ops = None

    # By default, all QueryObjects are available for non-logged in users.
    public = True

    def __init__(self, field_name=None, label=None, public=None):
        if field_name is not None:
            self.field_name = field_name

        if label is not None:
            self.label = label

        if self.label is None:
            self.label = get_verbose_name(self.field_name)

        if public is not None:
            self.public = public

    def value_from_web(self, value):
        """
        Prepare the value from web for use in self.real_query function.
        """
        return value

    def impacts_query(self, operator, value):
        """Returns True or False, depending on the operator and value.

        For some fields, combination of operator and value does not
        change the results. For example, if you have a string field,
        you have an empty value and you use operator EQUALS - this
        will find you every empty field in the database, which is
        OK. But, if you use operator CONTAINS, this query will not
        have any impact on the results, as looking doing
        SELECT ... WHERE field LIKE '%%' will not do anything.

        So, if given operator, value pair can impact query results,
        return True here.

        """

        # By default we return True here, feel free to override this
        # in subclasses.
        return True

    def value_for_description(self, value):
        """Return value for description - readable for end-user, for placement
        on web page."""
        return self.value_from_web(value)

    def value_to_web(self, value):
        return value

    def query_for(self, value, operation):
        return self.real_query(self.value_from_web(value), operation)

    def real_query(self, value, operation, validate_operation=True):
        """
        Prepare a real query - return a Q object.

        :param value: value which will be used in the query
        :param validate_operation: if operation is not found in
        EQUALITY_OPS_ALL, raise UnknownOperation.
        :ptype validate_operation: bool
        :rtype: django.db.models.Q
        """
        ret = None

        if operation in [text(x) for x in EQUALITY_OPS_ALL]:
            ret = Q(**{self.field_name: value})

        else:
            if validate_operation:
                raise UnknownOperation(operation)

        if operation in DIFFERENT_ALL:
            return ~ret

        return ret


class StringQueryObject(QueryObject):
    type = STRING
    ops = STRING_OPS
    empty_value_description = _("(empty)")

    def impacts_query(self, operator, value):
        if operator in [CONTAINS, NOT_CONTAINS, STARTS_WITH, NOT_STARTS_WITH] \
                and not value:
            return False
        return True

    def value_from_web(self, value):
        if type(value) == str or type(value) == unicode:
            return value
        if type(value) == bytes:
            return value.decode("utf-8")

    def value_for_description(self, value):
        value = super(StringQueryObject, self).value_for_description(value)
        if not value:
            return self.empty_value_description
        return u'"%s"' % html.escape(value)

    def real_query(self, value, operation):
        ret = QueryObject.real_query(
            self, value, operation, validate_operation=False)

        if ret is not None:
            return ret

        elif operation in [text(x) for x in [CONTAINS, NOT_CONTAINS]]:
            ret = Q(**{self.field_name + "__icontains": value})

        elif operation in [text(x) for x in [STARTS_WITH, NOT_STARTS_WITH]]:
            ret = Q(**{self.field_name + "__startswith": value})

        else:
            raise UnknownOperation(operation)

        if operation in [NOT_CONTAINS, NOT_STARTS_WITH]:
            return ~ret

        return ret


class AutocompleteQueryObject(QueryObject):
    type = AUTOCOMPLETE
    ops = EQUALITY_OPS_MALE
    model = None
    url = None

    def __init__(
            self, field_name=None, label=None, model=None, url=None,
            public=None):
        super(AutocompleteQueryObject, self).__init__(
            field_name, label, public=public)

        if model is not None:
            self.model = model
            self.url = '/multiseek/autocomplete/' + str(model.__name__)

        if url is not None:
            self.url = url

    def get_url(self):
        if self.url:
            return self.url
        return '/multiseek/autocomplete/%s/' % self.model.__name__

    @classmethod
    def get_label(cls, model):
        return text(model)

    def value_from_web(self, value):
        # The value should be an integer:
        try:
            value = int(value)
        except (TypeError, ValueError):
            return

        try:
            value = self.model.objects.get(pk=value)
        except self.model.DoesNotExist:
            return

        return value

    def value_to_web(self, value):
        try:
            model = self.model.objects.get(pk=value)
        except self.model.DoesNotExist:
            return json.dumps([None, ''])
        return json.dumps([value, self.get_label(model)])

    def get_autocomplete_query(self, data):
        """This function should return an iterable, like a QuerySet. This
         iterable, in turn, will be used by JQuery UI widget on the web.

        :param data: string passed from web request.
        """

        def args(fld, elem):
            return {fld + "__icontains": elem}

        if data:
            # split by comma, space, etc.
            data = data.split(" ")

            ret = Q(**args(self.search_fields[0], data[0]))
            for f, v in zip(self.search_fields[1:], data[1:]):
                ret = ret & Q(**args(f, v))
            return self.model.objects.filter(ret)

        return self.model.objects.all()

    def get_autocomplete_label(self, elem):
        """This function returns a label for the elem, this label in turn
        will be used by JQuery UI widget."""
        return text(elem)


class DateQueryObject(QueryObject):
    type = DATE
    ops = DATE_OPS

    def value_from_web(self, value):
        value = json.loads(value)
        if len(value) == 1:
            if not value[0]:
                # assume 'today'
                return datetime.now().date()
            return parse(value[0]).date()

        ret = []
        for elem in value:
            if not elem:
                # assume 'today'
                ret.append(datetime.now().date())
                continue
            ret.append(parse(elem).date())
        return ret

    def real_query(self, value, operation):
        if operation in RANGE_OPS:
            ret = Q(**{self.field_name + '__gte': value[0],
                       self.field_name + '__lte': value[1]})
        elif operation in EQUALITY_OPS_ALL:
            return Q(**{self.field_name + "__range": (value, value + timedelta(days=1))})
        elif operation in DIFFERENT_ALL:
            return ~Q(**{self.field_name + "__range": (value, value + timedelta(days=1))})
        elif operation in GREATER_OPS_ALL:
            return Q(**{self.field_name + "__gt": value})
        elif operation in LESSER_OPS_ALL:
            return Q(**{self.field_name + "__lt": value})
        elif operation in GREATER_OR_EQUAL_OPS_ALL:
            return Q(**{self.field_name + "__gte": value})
        elif operation in LESSER_OR_EQUAL_OPS_ALL:
            return Q(**{self.field_name + "__lte": value})
        else:
            raise UnknownOperation(operation)

        if operation == RANGE_OPS[1]:
            return ~ret

        return ret


class RangeQueryObject(QueryObject):
    type = RANGE
    ops = RANGE_OPS

    def value_from_web(self, value):
        # value should be a list of integers

        value = json.loads(value)

        try:
            int(value[0])
            int(value[1])
        except (ValueError, TypeError, IndexError):
            return

        if len(value) > 2:
            return

        return [int(value[0]), int(value[1])]

    def real_query(self, value, operation):

        if value is None:
            return

        if operation in RANGE_OPS:
            ret = Q(**{self.field_name + '__gte': value[0],
                       self.field_name + '__lte': value[1]})
        else:
            raise UnknownOperation(operation)

        if operation == RANGE_OPS[1]:
            return ~ret

        return ret


class AbstractNumberQueryObject(QueryObject):
    ops = [EQUAL, DIFFERENT, GREATER, LESSER,
           GREATER_OR_EQUAL, LESSER_OR_EQUAL]

    def impacts_query(self, operator, value):
        if value is None and operator in [GREATER, LESSER,
                                          GREATER_OR_EQUAL,
                                          LESSER_OR_EQUAL]:
            return False
        return True

    def real_query(self, value, operation):
        if operation in [text(x) for x in EQUALITY_OPS_ALL]:
            return Q(**{self.field_name: value})
        elif operation in [text(x) for x in DIFFERENT_ALL]:
            return ~Q(**{self.field_name: value})
        elif operation in [text(x) for x in GREATER_OPS_ALL]:
            return Q(**{self.field_name + "__gt": value})
        elif operation in [text(x) for x in LESSER_OPS_ALL]:
            return Q(**{self.field_name + "__lt": value})
        elif operation in [text(x) for x in GREATER_OR_EQUAL_OPS_ALL]:
            return Q(**{self.field_name + "__gte": value})
        elif operation in [text(x) for x in LESSER_OR_EQUAL_OPS_ALL]:
            return Q(**{self.field_name + "__lte": value})
        else:
            raise UnknownOperation(operation)


class IntegerQueryObject(AbstractNumberQueryObject):
    type = INTEGER

    def value_from_web(self, value):
        try:
            return int(value)
        except (ValueError, TypeError):
            return


class DecimalQueryObject(AbstractNumberQueryObject):
    type = DECIMAL

    def value_from_web(self, value):
        try:
            return Decimal(value)
        except (TypeError, decimal.InvalidOperation):
            return



class ValueListQueryObject(QueryObject):
    type = VALUE_LIST
    ops = [EQUAL, DIFFERENT]
    values = None

    def __init__(self, field_name=None, label=None, values=None, public=None):
        super(ValueListQueryObject, self).__init__(
            field_name, label, public=public)
        if values is not None:
            self.values = values

BOOLEAN_TRUE_LABEL = _("yes")
BOOLEAN_FALSE_LABEL = _("no")

class BooleanQueryObject(QueryObject):
    type = VALUE_LIST
    ops = [EQUAL, DIFFERENT]

    true_label = BOOLEAN_TRUE_LABEL
    false_label = BOOLEAN_FALSE_LABEL

    values = [true_label, false_label]

    def value_from_web(self, value):
        if value == self.true_label:
            return True
        if value == self.false_label:
            return False
        return None

    def value_for_description(self, value):
        return value

Ordering = namedtuple("Ordering", ["field", "label"])


class ReportType(namedtuple("ReportType", "id label public")):
    def __new__(cls, id, label, public=True):
        return super(ReportType, cls).__new__(cls, id, label, public)


class _FrameInfo:
    frame = None
    field = None

    def __init__(self, frame=-1, field=0):
        self.frame = frame
        self.field = field

def get_ordering_key_name(no):
    key = "%s%s" % (MULTISEEK_ORDERING_PREFIX, no)
    key_dir = key + "_dir"
    return key, key_dir

class MultiseekRegistry:
    """This is a base class for multiseek registry. A registry is a list
    of registered fields, that will be used to render the multiseek form
    and to query the database.
    """

    model = None

    # This is a list of Ordering (namedtuples of field, label) with
    # information about fields, that can be used to sort results of your query.
    # Label will be used on web ui. Field name is the part, that gets passed
    # to queryset.order_by
    ordering = None
    order_boxes = [_("Sort by:"), _("then by"), _("then by")]

    report_types = None
    default_ordering = None

    def __init__(self):
        self.fields = []
        self.field_by_name = {}
        self.default_ordering = {}
        self.report_types = []

    def set_default_ordering(self, *args):
        self.default_ordering = {}
        ordering_dict = dict(
            (x.field, no) for no, x in enumerate(self.ordering))

        for no, elem in enumerate(args):
            key, key_dir = get_ordering_key_name(no)

            desc = False
            if elem.startswith("-"):
                desc = True
                elem = elem[1:]

            self.default_ordering[key] = ordering_dict[elem]
            if desc is True:
                self.default_ordering[key_dir] = "1"

    def get_fields(self, public=True):
        """Returns a list of fields, by default returning only public fields.
        """
        if public:
            return [x for x in self.fields if x.public]
        return self.fields

    def get_field_by_name(self, name):
        for field in self.fields:
            if text(field.label) == name:
                return field

    def add_field(self, field):
        """Add a field to multiseek registry.

        "ptype field: multiseek.logic.QueryObject
        """
        if field.field_name:
            for pfx in [MULTISEEK_ORDERING_PREFIX, MULTISEEK_REPORT_TYPE]:
                assert (not field.field_name.startswith(pfx)), \
                    "Field names cannot start with '" + pfx + "'"
        self.fields.append(field)
        self.field_by_name = dict([(f.label, f) for f in self.fields])

        # Check if every label is unique
        assert (len(self.field_by_name.keys()) == len(self.fields)), \
            "All fields must have unique names"

    def field_by_type(self, type, public=True):
        """Return a list of fields by type.
        """
        return [field for field in self.get_fields(public) if
                field.type == type]

    def extract(self, attr, public=True):
        """Extract an attribute out of every field.
        """
        return [getattr(field, attr) for field in self.get_fields(public)]

    def parse_field(self, field):
        """Parse a field (from JSON)

        :param field: dict containing 'field', 'operation' and 'value' elements.
        :type field: dict
        :returns: QueryObject
        :rtype: multiseek.logic.QueryObject subclass
        """
        # prev_op key is OPTIONAL
        for key in ['field', 'operator', 'value']:
            if key not in field:
                raise ParseError("Key %s not found in field %r" % (key, field))

        f = self.get_field_by_name(field['field'])
        if f is None:
            raise UnknownField("Field type %r not found!" % field)

        if field['operator'] not in [text(x) for x in f.ops]:
            raise UnknownOperation(
                "Operation %r not valid for field %r" % (
                    field['operator'], field['field']))

        if field.get('prev_op', None) not in [AND, OR, ANDNOT, None]:
            raise UnknownOperation("%r" % field)

        if f.impacts_query(field['value'], field['operator']):
            return f.query_for(field['value'], field['operator'])

    def get_query_recursive(self, data):
        """Recursivley get query, basing on a list of elements.
        """

        ret = None

        for elem in data[1:]:
            if type(elem) == list:
                qobj = self.get_query_recursive(elem)
                prev_op = elem[0]
            else:
                qobj = self.parse_field(elem)
                prev_op = elem.get('prev_op', None)

            if ret is None:
                ret = qobj
                continue

            if prev_op == AND:
                ret = ret & qobj
            elif prev_op == OR:
                ret = ret | qobj
            elif prev_op == ANDNOT:
                ret = ret & ~Q(qobj);
            else:
                raise UnknownOperation(
                    "%s not expected" % elem.get('prev_op', None))

        return ret

    def get_query(self, data):
        """Return a query for a given JSON.
        """
        return self.get_query_recursive(data)

    def get_report_types(self, only_public=False):
        if only_public:
            return [x for x in self.report_types if x.public]
        return self.report_types

    def get_report_type(self, data, only_public=False):
        default_retval = ''
        report_types = self.get_report_types(only_public=only_public)

        if report_types:
            default_retval = self.report_types[0].id

        if data is None or type(data) == list or 'report_type' not in data:
            return default_retval

        try:
            idx = int(data['report_type'])
        except (ValueError, IndexError, TypeError):
            return default_retval

        try:
            return report_types[idx].id
        except IndexError:
            return default_retval

    def get_query_for_model(self, data, removed_manually=None):
        if data is None:
            return self.model.objects.all()

        # Fix for pre-0.8 versions
        if type(data) != dict:
            data = {'form_data': data}

        if 'form_data' in data:
            query = self.get_query(data['form_data'])
            retval = self.model.objects.filter(query)
        else:
            retval = self.model.objects.all()

        if removed_manually:
            retval = retval.exclude(pk__in=removed_manually)

        sb = []

        ordering  = data.get("ordering")
        if ordering:
            for no, element in enumerate(self.order_boxes):
                key, key_dir = get_ordering_key_name(no)
                if key in ordering:
                    try:
                        sort_idx = int(ordering[key])
                    except (TypeError, ValueError):
                        continue

                    try:
                        srt = self.ordering[sort_idx].field
                    except (IndexError, TypeError, ValueError):
                        continue

                    if not srt:
                        continue

                    if key_dir in ordering and ordering[key_dir] == "1":
                        srt = "-" + srt

                    sb.append(srt)

            if sb:
                retval = retval.order_by(*sb)

        return retval

    def recreate_form_recursive(self, element, info):
        result = []
        info.frame += 1

        current_frame = info.frame

        for elem in element[1:]:
            if type(elem) == list:
                result.append(
                    "$('#frame-%s').multiseekFrame('addFrame', '%s')" % (
                        current_frame, elem[0]))
                result.extend(self.recreate_form_recursive(elem, info))

            else:
                if elem.get("prev_op", None) not in [AND, OR, ANDNOT, None]:
                    raise ParseError("prev_op = %r" % elem.get("prev_op", None))

                prev_op = elem.get('prev_op', None)
                if prev_op == None or prev_op == 'None':
                    prev_op = 'null';
                else:
                    prev_op = "'" + prev_op + "'"
                s = "$('#frame-%i').multiseekFrame('addField', '%s', '%s', '%s', %s)"
                value = self.get_field_by_name(elem['field']).value_to_web(
                    elem['value'])

                result.append(s % (
                    current_frame, elem['field'], elem['operator'], value,
                    prev_op))
                info.field += 1

        return result

    def recreate_form(self, data):
        """Recreate a JavaScript code to create a given form, basing
        on a list.

        :returns: Javascript code to embed on the multiseek form page, which
        will recreate form
        :rtype: safestr
        """

        info = _FrameInfo()
        result = []
        if type(data) != dict:
            raise ParseError

        if 'form_data' in data:
            result = self.recreate_form_recursive(data['form_data'], info)
        foundation = []

        ordering = data.get("ordering")
        if ordering is None:
            if self.default_ordering:
                ordering = self.default_ordering

        if ordering:
            for no, elem in enumerate(self.order_boxes):
                key = "%s%s" % (MULTISEEK_ORDERING_PREFIX, no)
                if key in ordering:
                    result.append(
                        '\t\t'
                        '$("select[name=%s] option").eq(%s).prop("selected", true)' % (
                            key, ordering[key]))
                    # foundation.append(
                    #     '\t\t\t'
                    #     'Foundation.libs.forms.refresh_custom_select($("select[name=%s]"), true)' % key
                    # )
                key = key + "_dir"
                if key in ordering:
                    if ordering[key] == "1":
                        result.append(
                            '\t\t'
                            '$("input[name=%s]").attr("checked", true)' % key)
                        foundation.append(
                            '\t\t\t'
                            '$("input[name=%s]").next().toggleClass("checked", true)' % key
                        )

        if 'report_type' in data:
            if data['report_type']:
                result.append(
                    '\t\t'
                    '$("select[name=%s] option").eq(%s).prop("selected", true)'
                    % (
                        MULTISEEK_REPORT_TYPE, data['report_type']))
                # foundation.append(
                #     '\t\t\t'
                #     'Foundation.libs.forms.refresh_custom_select($("select[name=%s]"), true)' % MULTISEEK_REPORT_TYPE
                # )

        ret = u";\n".join(result) + u";\n"
        if foundation:
            ret += u"\t\tif (window.Foundation) {\n" + u";\n".join(
                foundation) + u"\n\t\t}\n"
        return ret



def create_registry(model, *args, **kw):
    r = MultiseekRegistry()
    r.model = model
    for field in args:
        r.add_field(field)

    known_kwargs =['ordering', 'report_types']
    for arg in known_kwargs:
        if arg in kw:
            setattr(r, arg, kw.pop(arg))

    if 'default_ordering' in kw:
        r.set_default_ordering(*kw.pop("default_ordering"))
    if kw.keys():
        raise Exception("Unknown kwargs passed")
    return r


_cached_registry = {}

def get_registry(registry):
    """
    :rtype: MultiseekRegistry
    """

    if type(registry) is str or type(registry) is unicode:
        if registry not in _cached_registry:
            m = importlib.import_module(registry).registry
            _cached_registry[registry] = m
        return _cached_registry[registry]

    return registry



# -*- encoding: utf-8 -*-
from __future__ import unicode_literals
import json

from django.http.response import HttpResponse, Http404, HttpResponseServerError, \
    HttpResponseRedirect
from django.views.generic.base import View
from django import shortcuts, http
from django.core.urlresolvers import reverse
from django.db import transaction
from django.http import HttpResponseForbidden, HttpResponseNotFound
from django.views.generic import TemplateView, ListView
from django.utils.translation import ugettext_lazy as _, ugettext_lazy

from .logic import VALUE_LIST, AUTOCOMPLETE, AND, OR, get_registry, \
    UnknownOperation, ParseError, UnknownField, MULTISEEK_ORDERING_PREFIX
from multiseek.logic import MULTISEEK_REPORT_TYPE
from multiseek.models import SearchForm
from builtins import str as text


SAVED = 'saved'
OVERWRITE_PROMPT = 'overwrite-prompt'

ERR_FORM_NAME = _("Form name must not be empty")
ERR_LOADING_DATA = _("Error while loading form data")
ERR_PARSING_DATA = _("Error while parsing form data")
ERR_NO_FORM_DATA = _("No form data provided")

MULTISEEK_SESSION_KEY = 'multiseek_json'
MULTISEEK_SESSION_KEY_REMOVED = 'multiseek_json_removed'


def reverse_or_just_url(s):
    if s.startswith('/'):
        return s
    return reverse(s)


LAST_FIELD_REMOVE_MESSAGE = \
    _("The ability to remove the last field has been disabled.")


def user_allowed_to_save_forms(user):
    if hasattr(user, 'is_staff'):
        return user.is_staff


class MultiseekPageMixin:
    registry = None


class MultiseekFormPage(MultiseekPageMixin, TemplateView):
    """
    This view renders multiseek form and javascript required to manipulate
    it.
    """

    def get_context_data(self):
        registry = get_registry(self.registry)

        public = self.request.user.is_anonymous()

        fields = registry.get_fields(public)

        js_fields = json.dumps([text(x.label) for x in fields])
        js_ops = json.dumps(dict(
            [(text(f.label), [text(x) for x in f.ops]) for f in fields]))
        js_types = json.dumps(
            dict([(text(f.label), f.type) for f in fields]))

        js_autocompletes = json.dumps(
            dict([(text(field.label), reverse_or_just_url(field.get_url()))
                  for field in registry.field_by_type(
                    AUTOCOMPLETE, public)]))

        def get_values(values):
            if callable(values):
                return values()
            return values

        js_value_lists = json.dumps(
            dict([
                (text(field.label), [text(x) for x in get_values(field.values)])
                for field in registry.field_by_type(VALUE_LIST, public)]))

        initialize_empty_form = True
        form_data = self.request.session.get(MULTISEEK_SESSION_KEY, {})
        if form_data:
            form_data = json.loads(form_data)
            initialize_empty_form = False
        js_init = registry.recreate_form(form_data)

        js_removed = ",".join('"%s"' %x for x in self.request.session.get(MULTISEEK_SESSION_KEY_REMOVED, []))
        return dict(
            js_fields=js_fields, js_ops=js_ops, js_types=js_types,
            js_autocompletes=js_autocompletes, js_value_lists=js_value_lists,
            js_and=AND, js_or=OR, js_init=js_init,
            js_remove_message=LAST_FIELD_REMOVE_MESSAGE,
            js_removed=js_removed,
            user_allowed_to_save_forms=user_allowed_to_save_forms(
                self.request.user),
            initialize_empty_form=initialize_empty_form,
            order_boxes=registry.order_boxes,
            ordering=registry.ordering,
            report_types=registry.get_report_types(only_public=public),
            saved_forms=SearchForm.objects.get_for_user(self.request.user),
            MULTISEEK_ORDERING_PREFIX=MULTISEEK_ORDERING_PREFIX,
            MULTISEEK_REPORT_TYPE=MULTISEEK_REPORT_TYPE)


def reset_form(request):
    for key in [MULTISEEK_SESSION_KEY, MULTISEEK_SESSION_KEY_REMOVED]:
        if key in request.session:
            del request.session[key]
    return shortcuts.redirect("..")


def load_form(request, search_form_pk):
    try:
        sf = SearchForm.objects.get(pk=search_form_pk)
    except SearchForm.DoesNotExist:
        return HttpResponseNotFound()

    if request.user.is_anonymous() and not sf.public:
        return HttpResponseForbidden()

    request.session[MULTISEEK_SESSION_KEY] = sf.data
    return shortcuts.redirect("..")


class JSONResponseMixin(object):
    def render_to_response(self, context):
        return self.get_json_response(self.convert_context_to_json(context))

    def get_json_response(self, content, **httpresponse_kwargs):
        return http.HttpResponse(
            content, content_type='application/json', **httpresponse_kwargs)

    def convert_context_to_json(self, context):
        return json.dumps(context)


class MultiseekSaveForm(MultiseekPageMixin, JSONResponseMixin, TemplateView):
    def get(self, request, *args, **kw):
        if not user_allowed_to_save_forms(request.user):
            return HttpResponseForbidden()
        return super(MultiseekSaveForm, self).get(request, *args, **kw)

    post = get

    @transaction.atomic
    def get_context_data(self):
        _json = self.request.POST.get('json')
        name = self.request.POST.get("name")
        public = self.request.POST.get("public") == 'true'
        overwrite = self.request.POST.get('overwrite') == 'true'

        if not _json:
            return dict(result=text(ERR_NO_FORM_DATA))

        try:
            json.loads(_json)
        except ValueError:
            return dict(result=text(ERR_PARSING_DATA))

        try:
            get_registry(self.registry).recreate_form(json.loads(_json))
        except (TypeError, UnknownField, ParseError, UnknownOperation):
            return dict(result=text(ERR_LOADING_DATA))

        if not name:
            return dict(result=text(ERR_FORM_NAME))

        if SearchForm.objects.filter(name=name).count():
            if not overwrite:
                return dict(result=OVERWRITE_PROMPT)

            obj = SearchForm.objects.get(name=name)
            obj.public = public
            obj.data = _json
            obj.owner = self.request.user
            obj.save()

        else:
            obj = SearchForm.objects.create(
                name=name, public=public, data=_json, owner=self.request.user)

        return dict(result=SAVED, pk=obj.pk)


class MultiseekResults(MultiseekPageMixin, ListView):
    registry = None
    _json_cache = None

    def post(self, request, *args, **kwargs):
        if 'json' in request.POST:
            j = request.POST['json']
            session = request.session
            session[MULTISEEK_SESSION_KEY] = j
            session.save()
        return super(MultiseekResults, self).get(request, *args, **kwargs)

    def get_multiseek_data(self):
        if not self._json_cache:
            _json = self.request.session.get(MULTISEEK_SESSION_KEY)
            if _json is not None:
                self._json_cache = json.loads(_json)
            if self._json_cache is None:
                self._json_cache = {}
            if self._json_cache.get("ordering") is None:
                self._json_cache['ordering'] = get_registry(self.registry).default_ordering
        return self._json_cache

    def get_removed_records(self):
        return self.request.session.get(MULTISEEK_SESSION_KEY_REMOVED, [])

    def describe_multiseek_data(self):
        """Returns a string with a nicely-formatted query, so you can
        display the query to the user, in a results window, for example.
        """
        data = self.get_multiseek_data()
        registry = get_registry(self.registry)

        ugettext_lazy("andnot") # Leave this line.

        def _recur(d):
            cur = 0
            ret = ''

            while cur < len(d):

                if type(d[cur]) == list:
                    if d[cur][0] != None:
                        ret += ' <b>' + text(
                            ugettext_lazy(d[cur][0])).upper() + '</b> '
                    ret += '(' + _recur(d[cur][1:]) + ')'
                else:
                    f = registry.get_field_by_name(d[cur]['field'])

                    impacts_query = f.impacts_query(
                        d[cur]['operator'], d[cur]['value'])

                    value = f.value_for_description(d[cur]['value'])

                    if impacts_query:

                        if 'prev_op' in d[cur] and d[cur]['prev_op'] != None:
                            tmp = d[cur]['prev_op']
                            ret += ' <b>' + text(
                                ugettext_lazy(tmp)).upper() + '</b> '


                        ret += '%s %s %s' % (d[cur]['field'].lower(),
                                             d[cur]['operator'],
                                             value)

                cur += 1

            return ret



        if data is None:
            return ''
        if not data.get('form_data'):
            return ''

        return _recur(data['form_data'][1:])

    def get_context_data(self, **kwargs):
        public = self.request.user.is_anonymous()
        report_type = get_registry(self.registry) \
            .get_report_type(self.get_multiseek_data(),
                             only_public=public)
        description = self.describe_multiseek_data()
        removed_ids = self.get_removed_records()

        return super(ListView, self).get_context_data(
            report_type=report_type, description=description,
            removed_ids=removed_ids, **kwargs)

    def get_queryset(self):
        # TODO: jeżeli w sesji jest obiekt, którego NIE DA się sparse'ować, to wówczas błąd podnoś i to samo w klasie MultiseekFormPage
        return get_registry(self.registry).get_query_for_model(
            self.get_multiseek_data(),
            self.request.session.get(MULTISEEK_SESSION_KEY_REMOVED, []))


class MultiseekModelRouter(View):
    registry = None

    def get(self, request, model, *args, **kw):
        registry = get_registry(self.registry)
        for field in registry.fields:
            if field.type == AUTOCOMPLETE:
                if model == field.model.__name__:
                    return MultiseekModelAutocomplete(original=field).get(request)
        raise Http404


class MultiseekModelAutocomplete(View):
    # TODO: JSONResponseMixin
    qobj = None
    max_items = 10

    def get_queryset(self, request):
        return self.qobj.model.objects.all()

    def get(self, request, *args, **kwargs):
        q = request.GET.get('term', '').encode('utf-8')

        qset = self.original.get_autocomplete_query(q)

        ret = []
        for elem in qset[:self.max_items]:
            ret.append(
                {'id': elem.pk,
                 'label': self.original.get_autocomplete_label(elem),
                 'value': self.original.get_autocomplete_label(elem)})

        return HttpResponse(json.dumps(ret),
                            content_type='application/json')


JSON_OK = HttpResponse(json.dumps({'status':"OK"}), content_type='application/json')

def manually_add_or_remove(request, pk, add=True):
    data = request.session.get(MULTISEEK_SESSION_KEY_REMOVED, [])
    data = set(data)

    if add:
        if len(data) < 2048:
            data.add(pk)
        else:
            # Prevent DOS OOM attack
            return HttpResponseForbidden()

    else:
        try:
            data.remove(pk)
        except KeyError:
            pass

    request.session[MULTISEEK_SESSION_KEY_REMOVED] = list(data)

    return JSON_OK


def remove_by_hand(request, pk):
    """Add a record's PK to a list of manually removed records.

    User, via the web ui, can add or remove a record to a list of records
    removed "by hand". Those records will be explictly removed
    from the search results in the query function. The list of those
    records is cleaned when there is a form reset.
    """
    return manually_add_or_remove(request, pk)

def remove_from_removed_by_hand(request, pk):
    """Cancel manual record removal."""
    return manually_add_or_remove(request, pk, add=False)

def reenable_removed_by_hand(request):
    request.session[MULTISEEK_SESSION_KEY_REMOVED] = []
    return HttpResponseRedirect(request.META.get('HTTP_REFERER') or '..')



# -*- encoding: utf-8 -*-

import os
import sys

from setuptools import setup, Command
from distutils.command.build import build as _build
from setuptools.command.install_lib import install_lib as _install_lib
from distutils.cmd import Command


class compile_translations(Command):
    description = 'compile message catalogs to MO files via django compilemessages'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        curdir = os.getcwd()
        os.chdir(os.path.realpath('multiseek'))
        from django.core.management import call_command
        call_command('compilemessages')
        os.chdir(curdir)


class build(_build):
    sub_commands = [('compile_translations', None)] + _build.sub_commands


class install_lib(_install_lib):
    def run(self):
        self.run_command('compile_translations')
        _install_lib.run(self)



# Utility function to read the README file.
# Used for the long_description. It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


class RunTests(Command):
    description = "Run the django test suite from the testproj dir."

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        this_dir = os.getcwd()
        testproj_dir = os.path.join(this_dir, "test_project")
        os.chdir(testproj_dir)
        sys.path.append(testproj_dir)
        from django.core.management import execute_manager

        os.environ["DJANGO_SETTINGS_MODULE"] = 'test_project.settings'
        settings_file = os.environ["DJANGO_SETTINGS_MODULE"]
        settings_mod = __import__(settings_file, {}, {}, [''])
        execute_manager(settings_mod, argv=[
            __file__, "test", "multiseek", "--traceback"])
        os.chdir(this_dir)


if 'sdist' in sys.argv:
    # clear compiled mo files before building the distribution
    walk = os.walk(os.path.join(os.getcwd(), 'multiseek/locale'))
    for dirpath, dirnames, filenames in walk:
        if not filenames:
            continue

        for fn in ['django.mo', 'djangojs.mo']:
            if fn in filenames:
                os.unlink(os.path.join(dirpath, fn))

else:
    # Always try to build messages, fail if django not present
    # (I hate incomplete releases)
    dir = os.getcwd()
    os.chdir(os.path.join(dir, 'multiseek'))
    os.system('django-admin.py compilemessages')
    os.chdir(dir)

def reqs(f):
    for elem in open(f).readlines():
        elem = elem.strip()
        if elem.startswith("#") or elem.startswith("-r"):
            continue
        yield elem

        
setup(
    name='django-multiseek',
    version='0.9.24',
    description='Build a form to seek records using multiple parameters',
    author=u'Michał Pasternak',
    author_email='michal.dtz@gmail.com',
    url='http://TODO',
    packages=['multiseek'],
    package_data={'multiseek': [
        'locale/*/LC_MESSAGES/*',
        'static/multiseek/*.js',
        'static/multiseek/*.css',
        'templates/multiseek/*.html']},
    include_package_data=True,
    zip_safe=False,
    long_description=read("README.md"),
    license='MIT',
    keywords='django multiseek',
    install_requires=list(reqs("requirements.txt")),
    tests_require=list(reqs("requirements_dev.txt")),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    cmdclass={'build': build, 'install_lib': install_lib,
              'compile_translations': compile_translations,
              'test': RunTests},
)



# -*- encoding: utf-8 -*-
import json

from django.contrib.auth.models import AnonymousUser, User
from django.test import TestCase
from django.test.client import RequestFactory
from mock import MagicMock
from model_mommy import mommy

from multiseek.logic import create_registry, StringQueryObject, \
    ValueListQueryObject, AutocompleteQueryObject, EQUALITY_OPS_ALL, EQUAL
from multiseek.models import SearchForm
from multiseek.views import MultiseekFormPage, MULTISEEK_SESSION_KEY, \
    reset_form, get_registry, user_allowed_to_save_forms, MultiseekSaveForm, \
    ERR_NO_FORM_DATA, ERR_PARSING_DATA, ERR_LOADING_DATA, ERR_FORM_NAME, \
    OVERWRITE_PROMPT, SAVED, load_form, MultiseekResults
from test_app import multiseek_registry
from test_app.models import Author
from builtins import str as text


class Session(dict):
    def save(self):
        return None


def setup_anonymous_session(request):
    request.user = AnonymousUser()
    request.session = Session()
    return request


class RegistryMixin:

    def setUp(self):
        self.registry = create_registry(
            None,
            StringQueryObject('foo'),
            StringQueryObject('bar'),
            ValueListQueryObject(field_name='baz', values=['a', 'b', 'c']),
            AutocompleteQueryObject(field_name='quux', url='/LOL/', model=Author)
            )

        self.request = setup_anonymous_session(RequestFactory().get('/'))


class TestViews(RegistryMixin, TestCase):

    def test_new_page(self):
        mfp = MultiseekFormPage(registry=self.registry)
        mfp.request = self.request
        ret = mfp.get_context_data()

    def test_multiseek(self):
        self.request.session[MULTISEEK_SESSION_KEY] = json.dumps(
            {'form_data':
                 [None, dict(field="foo", prev_op="or", operator=text(EQUALITY_OPS_ALL[0]), value="foo")]})

        mfp = MultiseekFormPage(registry=self.registry)
        mfp.request = self.request

        ret = mfp.get_context_data()

        self.assertEquals(ret['js_fields'], '["foo", "bar", "baz", "quux"]')
        self.assertEquals(ret['js_autocompletes'], '{"quux": "/LOL/"}')
        self.assertEquals(ret['js_value_lists'], '{"baz": ["a", "b", "c"]}')
        self.assertEquals(
            ret['js_init'],
            u"$('#frame-0').multiseekFrame('addField', 'foo', 'equals', 'foo', 'or');\n")

    def test_reset_form(self):
        self.request.session[MULTISEEK_SESSION_KEY] = '123'
        ret = reset_form(self.request)

        self.assertEquals(self.request.session, {})

        # No error on subsequent calls
        reset_form(self.request)

    def test_get_registry(self):
        self.assertEquals(get_registry({}), {})

        self.assertEquals(
            get_registry('test_app.multiseek_registry'),
            multiseek_registry.registry)

    def test_user_allowed_to_save_forms(self):
        class MockUser:
            is_staff = True

        self.assertEquals(user_allowed_to_save_forms(MockUser), MockUser.is_staff)
        MockUser.is_staff = False
        self.assertEquals(user_allowed_to_save_forms(MockUser), MockUser.is_staff)

        class MockUserWithout:
            pass

        self.assertEquals(user_allowed_to_save_forms(MockUserWithout), None)

class TestMultiseekSaveForm(RegistryMixin, TestCase):

    def setUp(self):
        RegistryMixin.setUp(self)
        self.msp = MultiseekSaveForm(registry=self.registry)
        self.msp.request = self.request

    def test_save_form_anon_user(self):
        res = self.msp.post(self.request)
        self.assertEquals(res.status_code, 403)

        self.assertEquals(self.msp.post, self.msp.get)

    def test_get_context_data(self):
        self.request.POST = {}
        
        self.request.POST['json'] = None
        self.assertEquals(
            self.msp.get_context_data(), 
            dict(result=text(ERR_NO_FORM_DATA)))
        
        self.request.POST['json'] = "wcale, nie, json"
        self.assertEquals(
            self.msp.get_context_data(),
            dict(result=text(ERR_PARSING_DATA)))
        
        self.request.POST['json'] = '[{"field": "foo", "bad": "field"}]'
        self.assertEquals(
            self.msp.get_context_data(),
            dict(result=text(ERR_LOADING_DATA)))
        
        self.request.POST['json'] = \
            '{"form_data": [{"field": "foo", "operation": "' \
            + text(EQUAL) \
            + '", "value": "foo"}]}'
        self.request.POST['name'] = ''
        self.assertEquals(
            self.msp.get_context_data(),
            dict(result=text(ERR_FORM_NAME)))

        sf = mommy.make(SearchForm, name='foo')
        self.request.POST['name'] = 'foo'
        self.assertEquals(
            self.msp.get_context_data(),
            dict(result=OVERWRITE_PROMPT))

        self.request.POST['overwrite'] = 'true'
        self.request.user = mommy.make(User)
        self.assertEquals(
            self.msp.get_context_data()['result'],
            'saved') # dict(result=SAVED, pk=1))

        self.assertEquals(SearchForm.objects.all().count(), 1)
        self.assertEquals(SearchForm.objects.all()[0].public, False)

        self.request.POST['public'] = 'true'
        self.msp.get_context_data()
        self.assertEquals(SearchForm.objects.all()[0].public, True)


class TestMultiseekLoadForm(TestCase):
    def setUp(self):
        self.user = User.objects.create_user('test', 'test@foo.pl', 'test')

        self.anon_req = setup_anonymous_session(RequestFactory().get('/'))

        self.normal_req = RequestFactory().get('/')
        self.normal_req.user = self.user
        self.normal_req.session = Session()

    def test_load_form_unexistent(self):
        res = load_form(self.anon_req, 1)
        self.assertEquals(res.status_code, 404)

    def test_load_form_existent_public_ok(self):
        sf = SearchForm.objects.create(
            name='foo', owner=self.user, public=True, data='some data')
        res = load_form(self.anon_req, sf.pk)
        self.assertEquals(res.status_code, 302)
        self.assertEquals(
            self.anon_req.session[MULTISEEK_SESSION_KEY],
            sf.data)

    def test_load_form_forbidden(self):
        sf = SearchForm.objects.create(
            name='foo', owner=self.user, public=False, data='some data')
        res = load_form(self.anon_req, sf.pk)
        self.assertEquals(res.status_code, 403)

    def test_load_form_non_public_logged_in_user(self):
        sf = SearchForm.objects.create(
            name='foo', owner=self.user, public=False, data='some data')
        res = load_form(self.normal_req, sf.pk)
        self.assertEquals(res.status_code, 302)


class TestMultiseekResults(RegistryMixin, TestCase):

    def setUp(self):
        RegistryMixin.setUp(self)
        self.registry.model = MagicMock()
        self.mr = MultiseekResults(registry=self.registry)
        self.mr.request = self.request
        self.request.session[MULTISEEK_SESSION_KEY] = json.dumps(
            {'form_data': [{'field': text(self.registry.fields[0].label),
              'operation': text(self.registry.fields[0].ops[0]),
              'value': u'foobar'}]})

    def test_post(self):
        res = self.mr.post(self.request)
        self.assertEquals(res.status_code, 200)

    def test_get_queryset(self):
        res = self.mr.get_queryset()



# -*- encoding: utf-8 -*-

from django.contrib.auth.models import AnonymousUser, User
from django.test import TransactionTestCase
from multiseek.models import SearchForm
from model_mommy import mommy


class TestModels(TransactionTestCase):
    def test_search_form_manager(self):
        u = mommy.make(User)

        s1 = mommy.make(SearchForm, owner=u, public=False, name='A')
        s2 = mommy.make(SearchForm, owner=u, public=True, name='B')

        res = SearchForm.objects.get_for_user(AnonymousUser())
        self.assertEquals(list(res), [s2])

        res = SearchForm.objects.get_for_user(u)
        self.assertEquals(list(res), [s1, s2])